import numpy as np
from numba import jit, njit

from hystore.core import validation as val

DEFAULT_CHUNKSIZE = 1 << 20
INVALID_INDEX = 1 << 62


def chunks(length, chunksize=1 << 20):
    cur = 0
    while cur < length:
        next_ = min(length, cur + chunksize)
        yield cur, next_
        cur = next_


@njit
def safe_map(data_field, map_field, map_filter, empty_value):
    result = np.zeros_like(map_field, dtype=data_field.dtype)
    for i in range(len(map_field)):
        if map_filter[i]:
            result[i] = data_field[map_field[i]]
        else:
            result[i] = empty_value
    return result


@njit
def map_valid(data_field, map_field, result=None):
    if result is None:
        result = np.zeros_like(map_field, dtype=data_field.dtype)
    for i in range(len(map_field)):
        if map_field[i] < INVALID_INDEX:
            result[i] = data_field[map_field[i]]
    return result


def ordered_map_valid_stream(data_field, map_field, result_field, chunksize=DEFAULT_CHUNKSIZE):
    df_it = iter(chunks(len(data_field.data)))
    mf_it = iter(chunks(len(map_field.data)))
    df_range = next(df_it)
    mf_range = next(mf_it)
    dfc = data_field.data[df_range[0]:df_range[1]]
    mfc = map_field.data[mf_range[0]:mf_range[1]]
    rslt = np.zeros(chunksize, dtype=data_field.data.dtype)

    i = 0
    while i < len(map_field.data):
        ii, d = ordered_map_valid_partial(dfc, mfc, rslt)



# 0 2 3 4 5 7 8 9 11 12 14 15 17 18 19
# 0 0 0 0 1 1 1 1  2  2  2  2  3  3  3
def ordered_map_valid_partial(data_field, map_field, result):
    i = 0
    while i < len(map_field):
        val = map_field[i]
        if val < INVALID_INDEX:
            if val >= len(data_field):
                # need a new data_field chunk
                return i, val
            result[i] = val
        i += 1
    return i, len(data_field)





@njit
def data_iterator(data_field, chunksize=1 << 20):
    cur = np.int64(0)
    chunks_ = chunks(len(data_field.data), chunksize)
    for c in chunks_:
        start = c[0]
        data = data_field.data[start:start + chunksize * 2]
        for v in range(c[0], c[1]):
            yield data[v]

@njit
def apply_filter_to_index_values(index_filter, indices, values):
    # pass 1 - determine the destination lengths
    cur_ = indices[:-1]
    next_ = indices[1:]
    count = 0
    total = 0
    for i in range(len(index_filter)):
        if index_filter[i] == True:
            count += 1
            total += next_[i] - cur_[i]
    dest_indices = np.zeros(count+1, indices.dtype)
    dest_values = np.zeros(total, values.dtype)
    dest_indices[0] = 0
    count = 1
    total = 0
    for i in range(len(index_filter)):
        if index_filter[i] == True:
            n = next_[i]
            c = cur_[i]
            delta = n - c
            dest_values[total:total + delta] = values[c:n]
            total += delta
            dest_indices[count] = total
            count += 1
    return dest_indices, dest_values


@njit
def apply_indices_to_index_values(indices_to_apply, indices, values):
    # pass 1 - determine the destination lengths
    cur_ = indices[:-1]
    next_ = indices[1:]
    count = 0
    total = 0
    for i in indices_to_apply:
        count += 1
        total += next_[i] - cur_[i]
    dest_indices = np.zeros(count+1, indices.dtype)
    dest_values = np.zeros(total, values.dtype)
    dest_indices[0] = 0
    count = 1
    total = 0
    for i in indices_to_apply:
        n = next_[i]
        c = cur_[i]
        delta = n - c
        dest_values[total:total + delta] = values[c:n]
        total += delta
        dest_indices[count] = total
        count += 1
    return dest_indices, dest_values


@njit
def apply_spans_index_of_min(spans, src_array, dest_array):
    for i in range(len(spans)-1):
        cur = spans[i]
        next = spans[i+1]

        if next - cur == 1:
            dest_array[i] = cur
        else:
            dest_array[i] = cur + src_array[cur:next].argmin()

    return dest_array


@njit
def apply_spans_index_of_max(spans, src_array, dest_array):
    for i in range(len(spans)-1):
        cur = spans[i]
        next = spans[i+1]

        if next - cur == 1:
            dest_array[i] = cur
        else:
            dest_array[i] = cur + src_array[cur:next].argmax()

    return dest_array


@njit
def apply_spans_index_of_first(spans, dest_array):
    dest_array[:] = spans[:-1]


@njit
def apply_spans_index_of_last(spans, dest_array):
    dest_array[:] = spans[1:] - 1


@njit
def apply_spans_index_of_min_filter(spans, src_array, dest_array, filter_array):
    for i in range(len(spans) - 1):
        cur = spans[i]
        next = spans[i + 1]
        if next - cur == 0:
            filter_array[i] = False
        elif next - cur == 1:
            filter_array[i] = True
            dest_array[i] = cur
        else:
            filter_array[i] = True
            dest_array[i] = cur + src_array[cur:next].argmin()

    return dest_array, filter_array


@njit
def apply_spans_index_of_max_filter(spans, src_array, dest_array, filter_array):
    for i in range(len(spans) - 1):
        cur = spans[i]
        next = spans[i + 1]
        if next - cur == 0:
            filter_array[i] = False
        elif next - cur == 1:
            filter_array[i] = True
            dest_array[i] = cur
        else:
            filter_array[i] = True
            dest_array[i] = cur + src_array[cur:next].argmax()

    return dest_array, filter_array


@njit
def apply_spans_index_of_first_filter(spans, dest_array, filter_array):
    for i in range(len(spans) - 1):
        cur = spans[i]
        next = spans[i + 1]
        if next - cur == 0:
            filter_array[i] = False
        else:
            filter_array[i] = True
            dest_array[i] = spans[i]

    return dest_array, filter_array


@njit
def apply_spans_index_of_last_filter(spans, dest_array, filter_array):
    for i in range(len(spans) - 1):
        cur = spans[i]
        next = spans[i + 1]
        if next - cur == 0:
            filter_array[i] = False
        else:
            filter_array[i] = True
            dest_array[i] = spans[i+1]-1

    return dest_array, filter_array


@njit
def apply_spans_count(spans, dest_array):
    for i in range(len(spans)-1):
        dest_array[i] = np.int64(spans[i+1] - spans[i])


@njit
def apply_spans_first(spans, src_array, dest_array):
    dest_array[:] = src_array[spans[:-1]]


@njit
def apply_spans_last(spans, src_array, dest_array):
    spans = spans[1:]-1
    dest_array[:] = src_array[spans]


@njit
def apply_spans_max(spans, src_array, dest_array):

    for i in range(len(spans)-1):
        cur = spans[i]
        next = spans[i+1]
        if next - cur == 1:
            dest_array[i] = src_array[cur]
        else:
            dest_array[i] = src_array[cur:next].max()


@njit
def apply_spans_min(spans, src_array, dest_array):

    for i in range(len(spans)-1):
        cur = spans[i]
        next = spans[i+1]
        if next - cur == 1:
            dest_array[i] = src_array[cur]
        else:
            dest_array[i] = src_array[cur:next].min()


# def _apply_spans_concat(spans, src_field):
#     dest_values = [None] * (len(spans)-1)
#     for i in range(len(spans)-1):
#         cur = spans[i]
#         next = spans[i+1]
#         if next - cur == 1:
#             dest_values[i] = src_field[cur]
#         else:
#             src = [s for s in src_field[cur:next] if len(s) > 0]
#             if len(src) > 0:
#                 dest_values[i] = ','.join(utils.to_escaped(src))
#             else:
#                 dest_values[i] = ''
#             # if len(dest_values[i]) > 0:
#             #     print(dest_values[i])
#     return dest_values


@njit
def apply_spans_concat(spans, src_index, src_values, dest_index, dest_values,
                       max_index_i, max_value_i, s_start):
    separator = np.frombuffer(b',', dtype=np.uint8)[0]
    delimiter = np.frombuffer(b'"', dtype=np.uint8)[0]
    if s_start == 0:
        index_i = np.uint32(1)
        index_v = np.int64(0)
        dest_index[0] = spans[0]
    else:
        index_i = np.uint32(0)
        index_v = np.int64(0)

    s_end = len(spans)-1
    for s in range(s_start, s_end):
        cur = spans[s]
        next = spans[s+1]
        cur_src_i = src_index[cur]
        next_src_i = src_index[next]

        dest_index[index_i] = next_src_i
        index_i += 1

        if next_src_i - cur_src_i > 1:
            if next - cur == 1:
                # only one entry to be copied, so commas not required
                next_index_v = next_src_i - cur_src_i + np.int64(index_v)
                dest_values[index_v:next_index_v] = src_values[cur_src_i:next_src_i]
                index_v = next_index_v
            else:
                # check to see how many non-zero-length entries there are; >1 means we must
                # separate them by commas
                non_empties = 0
                for e in range(cur, next):
                   if src_index[e] < src_index[e+1]:
                       non_empties += 1
                if non_empties == 1:
                    # only one non-empty entry to be copied, so commas not required
                    next_index_v = next_src_i - cur_src_i + np.int64(index_v)
                    dest_values[index_v:next_index_v] = src_values[cur_src_i:next_src_i]
                    index_v = next_index_v
                else:
                    # the outer conditional already determines that we have a non-empty entry
                    # so there must be multiple non-empty entries and commas are required
                    for e in range(cur, next):
                        src_start = src_index[e]
                        src_end = src_index[e+1]
                        comma = False
                        quotes = False
                        for i_c in range(src_start, src_end):
                            if src_values[i_c] == separator:
                                comma = True
                            elif src_values[i_c] == delimiter:
                                quotes = True

                        d_index = np.int64(0)
                        if comma or quotes:
                            dest_values[d_index] = delimiter
                            d_index += 1
                            for i_c in range(src_start, src_end):
                                if src_values[i_c] == delimiter:
                                    dest_values[d_index] = src_values[i_c]
                                    d_index += 1
                                dest_values[d_index] = src_values[i_c]
                                d_index += 1
                            dest_values[d_index] = delimiter
                            d_index += 1
                        else:
                            s_len = np.int64(src_end - src_start)
                            dest_values[index_v:index_v + s_len] = src_values[src_start:src_end]
                            d_index += s_len
                        index_v += np.int64(d_index)

        # if either the index or values are past the threshold, write them
        if index_i >= max_index_i or index_v >= max_value_i:
            break
    return s+1, index_i, index_v


def ordered_map_to_right_left_unique_streamed(left, right, left_to_right, chunksize=1 << 20):
    i = 0
    j = 0
    lc_it = iter(chunks(len(left.data), chunksize))
    lc_range = next(lc_it)
    rc_it = iter(chunks(len(right.data), chunksize))
    rc_range = next(rc_it)
    lc = left.data[lc_range[0]:lc_range[1]]
    rc = right.data[rc_range[0]:rc_range[1]]
    acc_written = 0

    ltri = np.zeros(chunksize, dtype=left_to_right.data.dtype)
    unmapped = 0
    while i < len(left.data) and j < len(right.data):
        ii, jj, u = ordered_map_to_right_left_unique_partial(i, lc, rc, ltri)
        unmapped += u
        if jj > 0:
            if val.is_field_parameter(left_to_right):
                left_to_right.data.write(ltri[:jj])
            else:
                left_to_right[acc_written:acc_written + jj] = ltri[:jj]
                acc_written += jj
        i += ii
        j += jj
        if i > lc_range[1]:
            raise ValueError("'i' has got ahead of current chunk; this shouldn't happen")
        if j > rc_range[1]:
            raise ValueError("'j' has got ahead of current chunk; this shouldn't happen")
        if i == lc_range[1] and i < len(left.data):
            lc_range = next(lc_it)
            lc = left.data[lc_range[0]:lc_range[1]]
        else:
            lc = lc[ii:]
        if j == rc_range[1] and j < len(right.data):
            rc_range = next(rc_it)
            rc = right.data[rc_range[0]:rc_range[1]]
        else:
            rc = rc[jj:]
    return unmapped > 0

"""
Returns:
[0]: how many positions forward i moved
[1]: how many positions forward j moved
[2]: how many elements were written
"""
@njit
def ordered_map_to_right_left_unique_partial(d_i, left, right, left_to_right):
    i = 0
    j = 0
    unmapped = 0
    while i < len(left) and j < len(right):
        if left[i] < right[j]:
            i += 1
        elif left[i] > right[j]:
            left_to_right[j] = INVALID_INDEX
            j += 1
            unmapped += 1
        else:
            left_to_right[j] = i + d_i
            if j+1 < len(right) and right[j+1] != right[j]:
                i += 1
            j += 1
    return i, j, unmapped


@njit
def ordered_map_to_right_left_unique(first, second, result):
    if len(second) != len(result):
        msg = "'second' and 'result' must be the same length"
        raise ValueError(msg)
    i = 0
    j = 0
    unmapped = 0
    while i < len(first) and j < len(second):
        if first[i] < second[j]:
            i += 1
        elif first[i] > second[j]:
            result[j] = INVALID_INDEX
            j += 1
            unmapped += 1
        else:
            result[j] = i
            if j+1 < len(second) and second[j+1] != second[j]:
                i += 1
            j += 1

    while j < len(second):
        result[j] = INVALID_INDEX
        j += 1

    return unmapped > 0


@njit
def ordered_map_to_right_both_unique(first, second, result):
    if len(second) != len(result):
        msg = "'second' and 'result' must be the same length"
        raise ValueError(msg)
    i = 0
    j = 0

    unmapped = 0
    while i < len(first) and j < len(second):
        if first[i] < second[j]:
            i += 1
        elif first[i] > second[j]:
            result[j] = INVALID_INDEX
            j += 1
            unmapped += 1
        else:
            result[j] = i
            i += 1
            j += 1

    while j < len(second):
        result[j] = INVALID_INDEX
        j += 1

    return unmapped > 0


@njit
def ordered_inner_map_result_size(left, right):
    i = 0
    j = 0
    result_size = 0

    while i < len(left) and j < len(right):
        if left[i] < right[j]:
            i += 1
        elif left[i] > right[j]:
            j += 1
        else:
            cur_i_count = 1
            while i+1 < len(left) and left[i + 1] == left[i]:
                cur_i_count += 1
                i += 1
            cur_j_count = 1
            while j+1 < len(right) and right[j + 1] == right[j]:
                cur_j_count += 1
                j += 1
            result_size += cur_i_count * cur_j_count
            i += 1
            j += 1
    return result_size


@njit
def ordered_inner_map_both_unique(left, right, left_to_inner, right_to_inner):
    i = 0
    j = 0
    cur_m = 0
    while i < len(left) and j < len(right):
        if left[i] < right[j]:
            i += 1
        elif left[i] > right[j]:
            j += 1
        else:
            left_to_inner[cur_m] = i
            right_to_inner[cur_m] = j
            cur_m += 1
            i += 1
            j += 1


def ordered_inner_map_left_unique_streamed(left, right, left_to_inner, right_to_inner,
                                           chunksize=1 << 20):
    i = 0
    j = 0
    left_chunks = chunks(len(left.data), 4)
    right_chunks = chunks(len(right.data), 4)

    lc_it = iter(left_chunks)
    lc_range = next(lc_it)
    rc_it = iter(right_chunks)
    rc_range = next(rc_it)
    lc = left.data[lc_range[0]:lc_range[1]]
    rc = right.data[rc_range[0]:rc_range[1]]

    lti = np.zeros(4, dtype=left_to_inner.data.dtype)
    rti = np.zeros(4, dtype=right_to_inner.data.dtype)
    while i < len(left.data) and j < len(right.data):
        ii, jj, m = ordered_inner_map_left_unique_partial(i, j, lc, rc, lti, rti)
        if m > 0:
            left_to_inner.data.write(lti[:m])
            right_to_inner.data.write(rti[:m])
        i += ii
        j += jj
        if i > lc_range[1]:
            raise ValueError("'i' has got ahead of current chunk; this shouldn't happen")
        if j > rc_range[1]:
            raise ValueError("'j' has got ahead of current chunk; this shouldn't happen")
        if i == lc_range[1] and i < len(left.data):
            lc_range = next(lc_it)
            lc = left.data[lc_range[0]:lc_range[1]]
        else:
            lc = lc[ii:]
        if j == rc_range[1] and j < len(right.data):
            rc_range = next(rc_it)
            rc = right.data[rc_range[0]:rc_range[1]]
        else:
            rc = rc[jj:]

"""
Returns:
[0]: how many positions forward i moved
[1]: how many positions forward j moved
[2]: how many elements were written
"""
@njit
def ordered_inner_map_left_unique_partial(d_i, d_j, left, right,
                                          left_to_inner, right_to_inner):
    i = 0
    j = 0
    m = 0
    while i < len(left) and j < len(right) and m < len(left_to_inner):
        if left[i] < right[j]:
            i += 1

        elif left[i] > right[j]:
            j += 1
        else:
            left_to_inner[m] = i + d_i
            right_to_inner[m] = j + d_j
            m += 1
            if j+1 < len(right) and right[j+1] != right[j]:
                i += 1
            j += 1
    return i, j, m


@njit
def ordered_inner_map_left_unique(left, right, left_to_inner, right_to_inner):
    i = 0
    j = 0
    cur_m = 0
    while i < len(left) and j < len(right):
        if left[i] < right[j]:
            i += 1
        elif left[i] > right[j]:
            j += 1
        else:
            cur_j = j
            while cur_j + 1 < len(right) and right[cur_j + 1] == right[cur_j]:
                cur_j += 1
            for jj in range(j, cur_j+1):
                left_to_inner[cur_m] = i
                right_to_inner[cur_m] = jj
                cur_m += 1
            i += 1
            j = cur_j + 1


@njit
def ordered_inner_map(left, right, left_to_inner, right_to_inner):
    i = 0
    j = 0
    cur_m = 0
    while i < len(left) and j < len(right):
        if left[i] < right[j]:
            i += 1
        elif left[i] > right[j]:
            j += 1
        else:
            cur_i = i
            while cur_i + 1 < len(left) and left[cur_i + 1] == left[cur_i]:
                cur_i += 1
            cur_j = j
            while cur_j + 1 < len(right) and right[cur_j + 1] == right[cur_j]:
                cur_j += 1
            for ii in range(i, cur_i+1):
                for jj in range(j, cur_j+1):
                    left_to_inner[cur_m] = ii
                    right_to_inner[cur_m] = jj
                    cur_m += 1
            i = cur_i + 1
            j = cur_j + 1



# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
import time
import numpy as np

from hystore.processing import numpy_buffer


class Dataset:
    """
    field_descriptors: a dictionary of field names to field descriptors that describe how the field
                       should be transformed when loading
    keys: a list of field names that represent the fields you wish to load and in what order they
          should be put. Leaving this blankloads all of the keys in csv column order
    """
    def __init__(self, source, field_descriptors=None, keys=None, filter_fn=None,
                 show_progress_every=False, start_from=None, stop_after=None, early_filter=None):
        self.names_ = list()
        self.fields_ = list()
        self.names_ = list()
        self.index_ = None

        csvf = csv.DictReader(source, delimiter=',', quotechar='"')
        available_keys = csvf.fieldnames

        if not keys:
            fields_to_use = available_keys
            index_map = [i for i in range(len(fields_to_use))]
        else:
            fields_to_use = keys
            index_map = [available_keys.index(k) for k in keys]

        early_key_index = None
        if early_filter is not None:
            if early_filter[0] not in available_keys:
                raise ValueError(
                    f"'early_filter': tuple element zero must be a key that is in the dataset")
            early_key_index = available_keys.index(early_filter[0])

        tstart = time.time()
        transforms_by_index = list()
        new_fields = list()

        # build a full list of transforms by index whether they are are being filtered by 'keys' or not
        for i_n, n in enumerate(available_keys):
            if field_descriptors and n in field_descriptors and\
               field_descriptors[n].strings_to_values and\
               field_descriptors[n].out_of_range_label is None:
                # transforms by csv field index
                transforms_by_index.append(field_descriptors[n])
            else:
                transforms_by_index.append(None)

        # build a new list of collections for every field that is to be loaded
        for i_n in index_map:
            if transforms_by_index[i_n] is not None:
                to_datatype = transforms_by_index[i_n].to_datatype
                if to_datatype == str:
                    new_fields.append(list())
                else:
                    new_fields.append(numpy_buffer.NumpyBuffer2(dtype=to_datatype))
            else:
                new_fields.append(list())

        # read the cvs rows into the fields
        csvf = csv.reader(source, delimiter=',', quotechar='"')
        ecsvf = iter(csvf)
        filtered_count = 0
        for i_r, row in enumerate(ecsvf):
            if show_progress_every:
                if i_r % show_progress_every == 0:
                    if filtered_count == i_r:
                        print(i_r)
                    else:
                        print(f"{i_r} ({filtered_count})")

            if start_from is not None and i_r < start_from:
                del row
                continue

            # TODO: decide whether True means filter or not filter consistently
            if early_filter is not None:
                if not early_filter[1](row[early_key_index]):
                    continue

            # TODO: decide whether True means filter or not filter consistently
            if not filter_fn or filter_fn(i_r):
                # for i_f, f in enumerate(fields):
                for i_df, i_f in enumerate(index_map):
                    f = row[i_f]
                    t = transforms_by_index[i_f]
                    try:
                        new_fields[i_df].append(f if not t else t.strings_to_values[f])
                    except Exception as e:
                        msg = "{}: key error for value {} (permitted values are {}"
                        print(msg.format(fields_to_use[i_f], f, t.strings_to_values))
                del row
                filtered_count += 1
                if stop_after and i_r >= stop_after:
                    break

        if show_progress_every:
            print(f"{i_r} ({filtered_count})")

        # assign the built sequences to fields_
        for i_f, f in enumerate(new_fields):
            if isinstance(f, list):
                self.fields_.append(f)
            else:
                self.fields_.append(f.finalise())
        self.index_ = np.asarray([i for i in range(len(self.fields_[0]))], dtype=np.uint32)
        self.names_ = fields_to_use
        print('loading took', time.time() - tstart, "seconds")

        #     if i > 0 and i % lines_per_dot == 0:
        #         if i % (lines_per_dot * newline_at) == 0:
        #             print(f'. {i}')
        #         else:
        #             print('.', end='')
        # if i % (lines_per_dot * newline_at) != 0:
        #     print(f' {i}')

    def sort(self, keys):
        #map names to indices
        if isinstance(keys, str):

            def single_index_sort(index):
                field = self.fields_[index]

                def inner_(r):
                    return field[r]

                return inner_
            self.index_ = sorted(self.index_,
                                 key=single_index_sort(self.field_to_index(keys)))
        else:

            kindices = [self.field_to_index(k) for k in keys]

            def index_sort(indices):
                def inner_(r):
                    t = tuple(self.fields_[i][r] for i in indices)
                    return t
                return inner_

            self.index_ = sorted(self.index_, key=index_sort(kindices))

        for i_f in range(len(self.fields_)):
            unsorted_field = self.fields_[i_f]
            self.fields_[i_f] = Dataset._apply_permutation(self.index_, unsorted_field)
            del unsorted_field

    @staticmethod
    def _apply_permutation(permutation, field):
        # n = len(permutation)
        # for i in range(0, n):
        #     print(i)
        #     pi = permutation[i]
        #     while pi < i:
        #         pi = permutation[pi]
        #     fields[i], fields[pi] = fields[pi], fields[i]
        # return fields
        if isinstance(field, list):
            sorted_field = [None] * len(field)
            for ip, p in enumerate(permutation):
                sorted_field[ip] = field[p]
        else:
            sorted_field = np.empty_like(field)
            for ip, p in enumerate(permutation):
                sorted_field[ip] = field[p]
        return sorted_field

    def field_by_name(self, field_name):
        return self.fields_[self.field_to_index(field_name)]

    def field_to_index(self, field_name):
        return self.names_.index(field_name)

    def value(self, row_index, field_index):
        return self.fields_[field_index][row_index]

    def value_from_fieldname(self, index, field_name):
        return self.fields_[self.field_to_index(field_name)][index]

    def row_count(self):
        return len(self.index_)

    def show(self):
        for ir, r in enumerate(self.names_):
            print(f'{ir}-{r}')



from datetime import datetime
import json
import csv
import pandas


from hystore.core import utils, persistence
from hystore.core import readerwriter as rw

def transform_from_reader_type(reader):
    if isinstance(reader, rw.FixedStringReader):
        return lambda x: x.decode()
    if isinstance(reader, rw.TimestampReader):
        return lambda x: datetime.fromtimestamp(x)
    return None


def schema_from_reader_type(reader):
    if isinstance(reader, rw.IndexedStringReader):
        return 'string'



def export_to_csv(destination, datastore, fields):

    expected_length = None
    for f in fields:
        if expected_length is None:
            expected_length = len(datastore.get_reader(f[1]))
        else:
            length = len(datastore.get_reader(f[1]))
            if length != expected_length:
                raise ValueError(f"field '{f[1]}' is not the same length as the first field:"
                                 f"expected {expected_length} but got {length}")
    print('expected_length:', expected_length)
    readers = [datastore.get_reader(f[1]) for f in fields]
    transforms = [transform_from_reader_type(r) for r in readers]

    with open(destination, 'w') as d:
        csvw = csv.writer(d)
        #header
        header = [None] * len(fields)
        names = [f[0] for f in fields]
        for i_f in range(len(fields)):
            header[i_f] = names[i_f]
        csvw.writerow(header)

        row = [None] * len(fields)
        for chunk in datastore.chunks(expected_length):
            s = slice(chunk[0], chunk[1])
            print(s)
            chunks = [r[s] for r in readers]
            for i_r in range(chunk[1] - chunk[0]):
                for i_c, c in enumerate(chunks):
                    if transforms[i_c] is not None:
                        row[i_c] = transforms[i_c](c[i_r])
                    else:
                        row[i_c] = c[i_r]
                csvw.writerow(row)


def export_schema(destination, readers):
    names = [r.name.split('/')[-1] for r in readers]
    schema_dict = {}
    for n in names:
        schema_dict[n] = []



# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

def iterate_over_patient_assessments(fields, filter_status, visitor):
    patient_ids = fields[1]
    cur_id = patient_ids[0]
    cur_start = 0
    cur_end = 0
    i = 1
    while i < len(filter_status):
        while patient_ids[i] == cur_id:
            cur_end = i
            i += 1
            if i >= len(filter_status):
                break

        visitor(fields, filter_status, cur_start, cur_end)

        if i < len(filter_status):
            cur_start = i
            cur_end = cur_start
            cur_id = patient_ids[i]


def iterate_over_patient_assessments2(patient_ids, filter_status, visitor):
    cur_id = patient_ids[0]
    cur_start = 0
    cur_end = 0
    i = 1
    while i < len(patient_ids):
        while patient_ids[i] == cur_id:
            cur_end = i
            i += 1
            if i >= len(patient_ids):
                break

        visitor(cur_id, filter_status, cur_start, cur_end)

        if i < len(patient_ids):
            cur_start = i
            cur_end = cur_start
            cur_id = patient_ids[i]



from collections import defaultdict

import h5py
import itertools
import scipy

from hystore.core.session import Session
from hystore.core import utils

class Result:
    def __init__(self):
        self.count = 0
        self.positive = 0
        self.negative = 0
        self.predicted_positive = 0
        self.predicted_negative = 0

    def update(self, old_tr, new_tr, prediction):
        self.count += 1
        if old_tr == 3 or new_tr == 4:
            self.positive += 1
        elif old_tr == 2 or new_tr == 3:
            self.negative += 1

        if prediction == 0:
            self.predicted_negative += 1
        else:
            self.predicted_positive += 1

    def __str__(self):
        return "count {}; neg {}; pos {}; -pred {}; +pred {}".format(
            self.count, self.negative, self.positive,
            self.predicted_negative, self.predicted_positive)


def supplement_paper_processing_2(ds, src_data):
    print(src_data['filtered_patients'].keys())
    print(src_data['final_assessments'].keys())
    ptnts = src_data['filtered_patients']
    asmts = src_data['final_assessments']
    supplement_fields = ('vs_multivitamins', 'vs_none', 'vs_omega_3',
                         'vs_vitamin_c', 'vs_vitamin_d', 'vs_zinc',)

    vs_none = ds.get(ptnts['vs_none']).data[:]
    print(utils.build_histogram(vs_none))
    vs_omega_3 = ds.get(ptnts['vs_omega_3']).data[:]
    print(utils.build_histogram(vs_omega_3))
    vs_vitamin_c = ds.get(ptnts['vs_vitamin_c']).data[:]
    print(utils.build_histogram(vs_vitamin_c))
    vs_vitamin_d = ds.get(ptnts['vs_vitamin_d']).data[:]
    print(utils.build_histogram(vs_vitamin_d))
    vs_zinc = ds.get(ptnts['vs_zinc']).data[:]
    print(utils.build_histogram(vs_zinc))
    vs_multivitamins = ds.get(ptnts['vs_multivitamins']).data[:]
    print(utils.build_histogram(vs_multivitamins))

    old_result = ds.get(ptnts['old_test_result']).data[:]
    print(utils.build_histogram(old_result))
    new_result = ds.get(ptnts['new_test_result']).data[:]
    print(utils.build_histogram(new_result))
    prediction = ds.get(asmts['prediction']).data[:]
    print(utils.build_histogram(prediction))

    vit_categories = defaultdict(Result)
    for i in range(len(vs_none)):
        key = (vs_none[i], vs_multivitamins[i], vs_omega_3[i],
               vs_vitamin_c[i], vs_vitamin_d[i], vs_zinc[i])
        vit_categories[key].update(old_result[i], new_result[i], prediction[i])

    results = sorted(list(vit_categories.items()))

    for k, v in results:
        print(k, v)
        print(k, v.positive / (v.negative + v.positive))

    all_positive = (old_result == 3) | (new_result == 4)
    all_negative = (all_positive == False) & ((old_result == 2) | (new_result == 3))
    all_positive_ratio = all_positive.sum() / (all_positive.sum() + all_negative.sum())
    print("overall tested:", all_positive_ratio)
    all_ppositive_ratio = prediction.sum() / len(prediction)
    print("overall predicted:", all_ppositive_ratio)
    neg = 0
    pos = 0
    pneg = 0
    ppos = 0
    for k, v in results:
        if k[0] == 1:
            neg += v.negative
            pos += v.positive
            pneg += v.predicted_negative
            ppos += v.predicted_positive
    pos_ratio = pos / (neg + pos)
    print('v_none tested:', pos_ratio, pos_ratio / all_positive_ratio)
    ppos_ratio = ppos / (pneg + ppos)
    print('v_none predicted:', ppos_ratio, ppos_ratio / all_ppositive_ratio)

    neg = 0
    pos = 0
    pneg = 0
    ppos = 0
    for k, v in results:
        if k[1] == 1:
            neg += v.negative
            pos += v.positive
            pneg += v.predicted_negative
            ppos += v.predicted_positive
    pos_ratio = pos / (neg + pos)
    print('v_multi tested:', pos_ratio, pos_ratio / all_positive_ratio)
    print('v_multi predicted:', ppos_ratio, ppos_ratio / all_ppositive_ratio)

    neg = 0
    pos = 0
    pneg = 0
    ppos = 0
    for k, v in results:
        if k[1] == 1 or k[3] == 1:
            neg += v.negative
            pos += v.positive
            pneg += v.predicted_negative
            ppos += v.predicted_positive
    pos_ratio = pos / (neg + pos)
    print('v_multi | v_vitamin_c tested:', pos_ratio, pos_ratio / all_positive_ratio)
    ppos_ratio = ppos / (pneg + ppos)
    print('v_multi | v_vitamin_c predicted:', ppos_ratio, ppos_ratio / all_ppositive_ratio)

    neg = 0
    pos = 0
    pneg = 0
    ppos = 0
    for k, v in results:
        if k[1] == 1 or k[4] == 1:
            neg += v.negative
            pos += v.positive
            pneg += v.predicted_negative
            ppos += v.predicted_positive
    pos_ratio = pos / (neg + pos)
    print('v_multi | v_vitamin_d tested:', pos_ratio, pos_ratio / all_positive_ratio)
    ppos_ratio = ppos / (pneg + ppos)
    print('v_multi | v_vitamin_d predicted:', ppos_ratio, ppos_ratio / all_ppositive_ratio)




if __name__ == '__main__':
    datastore = Session()
    src_file = '/home/ben/covid/ds_20200830_supplements.hdf5'
    with h5py.File(src_file, 'r') as src_data:
        supplement_paper_processing_2(datastore, src_data)
